#include <Servo.h>

/* Rauchmodul Sofware V1.0 (Stand 01.11.2021)
 * Ersteller P. Doggwiler
 * Datum: 29.10.2021
 * Über Dig. Eingang oder direkt ab Empfänger wird der Rauchgenerator eingeschaltet
 * Minimales PWM Signal am Ausgang steuert den Controller damit es auch im "Stillstand" raucht
 * Minimales PWM Signal über Trimmpoti einstellbar
 * Das PWM Signal vom Motor wird permanent eingelesen
 * => Signal > als 1500: Ausgabe PWM an Ausgang
 * => Signal < als 1500: Singal wird ugerechnet auf positiven Wert: Ausgabe =  3000 - Eingang
 * => Signal wird nur weitergeleitet wenn es grösser als das minimal eingestelle Signal ist
 * 
 * 
 * Alle Ausgangssignal werden bei der Initialisierung auf Wert 1500 gesetzt (Mittelstellung)
 * 
 * <<<<<<<<<<<<<<<<<<<<<<!!!!!!!! Start Wichtige Hinweise (02.08.2020) !!!!!!!!>>>>>>>>>>>>>>>>>>>>>>
 * Test haben ergeben, das bei gewissen Servos das PWM Signal beim Steuern das Signal auf den Emfpägerausgang überspringt, 
 * welches zum MulitSwitch Print führt. Dieses Siganlüberspringen wird eingelesen und Zählt den Counter hoch.
 * Einzige Lösung bisher ist der Austausch des Servos.
 * Das Problem konnte mangels Alternativen an Fernsteuerungen/Empfängern nur mit GRAUPNER HOTT Anlagen getestet und nachvollzogen werden.
 * <<<<<<<<<<<<<<<<<<<<<<!!!!!!!! Ende wichtige Hinweise !!!!!!!!>>>>>>>>>>>>>>>>>>>>>>
 * 
 */

 /*********************************************************************
  * Variablen welche angepasst werden müssen
  *********************************************************************/
int Minuten = 45;                               //Default Wert des maximal aktiven Ausgangs in Minuten



/*********************************************************
 * Start des Programms mit den Definierungen der Eingänge
 *********************************************************/
#define PinOut1 3           // Eingang für Freigabe Leistung mit digitalen Signal.
#define PinOut2 4           // Ausgangssignal für den Rauchgenerator Kontroller, das PWM Signal ist immer in "Vorwärtsrichtung"
#define PinOut3 5           // Ausgang für "Lastschaltung" => Am Besten über Arduino Relais
       


#define PinOutStart A0       //Blinkende LED im letzten Schritt der Setup Routine  => Ändern auf PIN ID A0 (Analog Output 0), ist die LED auf dem JoyIt Board
#define PinInA4 A4          //Auswahl für Run Modus ob Freigabe als Dig. Input (VCC) oder PWM Signal (GND)
#define PinInA5 A5          //Auswahl Programier oder Run Modus => VCC Programier- / GND Run Modus
#define PinInA6 A6          //AD Wandler: Einlesen Trimmptoi und Umwandlung in PWM Signal


//#define DEBUG             // Serial output setup
//#define DEBUG_PWM         // Serial output PWM Daten
//#define DEBUG_DIG         // Serial output Digitale Daten
//#define DEBUG_FRG         // Serial output Freigabe Leistung



//Servo Objekt Erstellen
Servo myservo;  // Servo-Objekt aus Library

 
volatile const int PinIn2 = 2;                        //Eingang für Aktivierung Rauchmodul über PWM Eingang ab Empfänger
volatile const int PinIn13 = 13;                      //Eingang für PWM Signal von Motor

unsigned long getPWM_Input = 0;                       //Eingelesendes PWM Signal
unsigned long getPWM_Output = 0;                      //PWM Ausgabe Signal
unsigned long setPWM_Minimal = 0;                     //Berechnetes Signal gemäss Programmierung über AD Wandler und Trimmer Signal

//Hilfsvariablen für Steuerung über Dig. Input oder PWM Input
bool setDigFreigabe = false;                          //False ist Freigabe Ausgang über PWM Signal Input an D2
bool setProgMod = false;                              //False ist im Run Modus => Eingang A05
bool setFreigabeStatus = false;                       //Toggle Status Signal damit bei der Abfrage nicht dauern der Ausgang geschaltet wird
unsigned int readProgWert = 0;                        //Hilfsvariable für einelsen des AD Wertes



//Zeit Variablen für Ausgang schalten
unsigned long alteZeit = 0;                            //Variable für Speicherung der alten Zeit für Auruf der Zeitkontrollfunkton alle 10 Sekunden
volatile unsigned long startZeit = 0;                  //Startzeit Aktivierung Ausgang
volatile unsigned long aktLaufzeit = 0;                //Total aktiver Leistungsausgang
volatile unsigned long maxTime = Minuten * 60000;       //Max Laufzeit des Leistungsausganges in Milisekunden umgerechnet
volatile bool setFreigabeLeistungTime = true;          //Variable für Freigabe des Leistungsausgangs

//Variable für PWM Signal
volatile unsigned long PWM_Start_Input = 1500;        //Servo ist auf Mitelstellung







void setup() 
{

  /**************************************************************
   * Wartezeit von 10s bis das Setup Programm ausgeführt wird.
   **************************************************************/
  long waitTime = millis();
  int WaitingTime = 10000;        //Wartezeit in Millisekunden => Default sind 10sek
  int waitTimeTemp;
  
  while(waitTimeTemp <= WaitingTime)
  {
    waitTimeTemp = millis() - waitTime;  
  }


  
  //Notwendig wenn die Ausgabe auf dem Seriellen Monitor für Debug Zwecke erfolgen soll
   //Serial.begin(115200);

   
   #ifdef DEBUG        // Serial output on 
   Serial.println("<<<<<<<<<<<<<<<<<<<< Start Setup Routine! >>>>>>>>>>>>>>>>>>>>>>");
   #endif


  pinMode(PinInA4, INPUT_PULLUP);              
  digitalWrite(PinInA4, LOW);
  
  pinMode(PinInA5, INPUT_PULLUP);              
  digitalWrite(PinInA5, LOW);                  

  pinMode(PinOutStart, OUTPUT);               
  digitalWrite(PinOutStart, LOW);             

  pinMode(PinOut3, OUTPUT);                   
  digitalWrite(PinOut3, LOW);                 
  
  pinMode(PinIn13, INPUT_PULLUP);             
  digitalWrite(PinIn13, LOW);               

  
  if(digitalRead(PinInA4) == HIGH)               //High = Freigabe Leistung über dig. Eingang => HIGH ist aktiv
  {
      pinMode(PinOut1, INPUT);
      digitalWrite(PinOut1, LOW);              //Counter bzw. Taster, Interner Pull up Widerstand auf 5V, auf 0V setzen
      setDigFreigabe = true;
   
       #ifdef DEBUG        // Serial output on 
       Serial.println("Freigabe Leistung über Dig. Ingput Signal");
       #endif 
  }
  else                                        //Freigabe Leistung über PWM Signa (min oder Max Ausschlag)
  {
     pinMode(PinIn2, INPUT_PULLUP); 
     digitalWrite(PinIn2, LOW);              //Counter bzw. Taster, Interner Pull up Widerstand auf 5V, auf 0V setzen
     setDigFreigabe = false;

     #ifdef DEBUG        // Serial output on 
     Serial.println("Freigabe Leistung über PWM Signal");
     #endif

  }


  if(digitalRead(PinInA5) == HIGH)
  {
    setProgMod = true;
    digitalWrite(PinOut3, HIGH);            //Leistungsausgang auf High setzen
  }

   #ifdef DEBUG        // Serial output on 
   Serial.print("SW Modues: ");
   if(setProgMod == true)
    Serial.println("Programmierung");
   else
    Serial.println("Run");

   
   #endif 

      //Servo Ausgänge auf Mittelstellung setzen
          myservo.attach(PinOut2);                      // Ausgang als Servo Signal setzen

      //AD Wert des Trimmpotis einlesen, wandeln und an PWM Minimal Variable zuweisen
        readProgWert = analogRead(PinInA6);
        setPWM_Minimal = map(readProgWert, 0, 1023, 1480, 1980);
        myservo.writeMicroseconds(setPWM_Minimal);


   #ifdef DEBUG        // Serial output on 
   Serial.println("Ausgabewert AD Wandlung beim Start:");
   Serial.print("Analog Read: ");
   Serial.print(readProgWert);
   Serial.print(" => ");
   Serial.print("Mapping Wert: ");
   Serial.println(setPWM_Minimal);   
   #endif




          /*****************************************************************************
          Start Kontrollausgabe nach dem Setup
          *****************************************************************************/
              
          digitalWrite(PinOutStart, HIGH);
          delay(1000);
          digitalWrite(PinOutStart, LOW);
          delay(1000);
          for(int i = 0; i < 5; i++)
          {
            digitalWrite(PinOutStart, HIGH);
            delay(200); 
            digitalWrite(PinOutStart, LOW);
            delay(200);
          }
        
          digitalWrite(PinOutStart, HIGH);
          delay(2000);
          digitalWrite(PinOutStart, LOW);




  #ifdef DEBUG        // Serial output on 
  Serial.println("<<<<<<<<<<<<<<<<<<<< Ende Setup Routine! >>>>>>>>>>>>>>>>>>>>>>");
  #endif
  /*****************************************************************************
  Ende Kontrollausgabe nach dem Setup
  *****************************************************************************/

  
}



void loop() {


      /********************************************
       * Programm Modus zum Einstellen des PWM Signals am Ausgangs
       ********************************************/

       if(setProgMod == true)
       {
        
        /*
        * setPWM_Minimal
        * analogRead(PinInA6) => Auflösung 0-1023
        * Aufruf Pgramm Funktion
        *map(value, fromLow, fromHigh, toLow, toHigh)
         *
         *  value =  Imputwert vom analogRead
         *  fromLow = 0;
         *  fromHigh = 1023; 
         *  toLow = 1480;
         *  toHigh = 1980;
         */

        readProgWert = analogRead(PinInA6);
        setPWM_Minimal = map(readProgWert, 0, 1023, 1480, 1980);
        myservo.writeMicroseconds(setPWM_Minimal);



         #ifdef DEBUG_PWM        // Serial output on 
         Serial.print("Analog Read: ");
         Serial.print(readProgWert);
         Serial.print(" => ");
         Serial.print("Mapping Wert: ");
         Serial.println(setPWM_Minimal);   
         #endif
           
        
       }

      /********************************************
       * Ende Programm Modus
       ********************************************/  








      /********************************************
       *Run Modus
       ********************************************/ 
      if(setProgMod == false && setFreigabeLeistungTime == true)
      {

          //Kontrolle ob die max. Laufzeit erreicht ist, Aufruf alle 10 sekunden
          if(  (millis()-alteZeit) > 10000)
          {   
              alteZeit = millis();
              LaufzeitKontrolle();
          }
    
           /********************************************
           * Einlesen Freigabe Leistung Ausgang
           ********************************************/
            if(setDigFreigabe == false)
            {
                 //PWM Signal ab Empfänger einlesen
                 getPWM_Input = pulseIn(PinIn2, HIGH);

                 #ifdef DEBUG_PWM        // Serial output on 
                 Serial.print("LE Ausgang PWM Aktivierung: ");
                 Serial.print(getPWM_Input);  
                 Serial.print(" | ");         
                 Serial.print("Status: ");
                 Serial.println(setFreigabeStatus); 
                 #endif
         
                  if((getPWM_Input < 1100 || getPWM_Input > 1700) && getPWM_Input > 10)
                  {
                      if(setFreigabeStatus == false)
                        setLeistungFreigabe(true);
                  }
                  else
                  {
                      if(setFreigabeStatus == true)
                        setLeistungFreigabe(false);
                  }
              
            }
            else
            {
    
                 #ifdef DEBUG_DIG        // Serial output on 

                 Serial.print("LE Ausgang Dig. Aktivierung: ");
                 Serial.print(digitalRead(PinOut1));  
                 Serial.print(" | ");         
                 Serial.print("Status: ");
                 Serial.println(setFreigabeStatus);
                
                 #endif          
        
                  if(digitalRead(PinOut1) == HIGH)
                  {
                      if(setFreigabeStatus == false)
                        setLeistungFreigabe(true);
                  }
                  else if(digitalRead(PinOut1) == LOW)
                  {
                      if(setFreigabeStatus == true)
                         setLeistungFreigabe(false);    
                  }
              
            }


          
           /*************************************
           * Weiterleiten Servo Eingang ans Ausgang
           * PWM > 1500 => Direkte Ausgabe
           * PWM < 1500 => 3000 - PWM Input Signal an Ausgang
           * setPWM_Minimal => Berechneter minimaler Wert über Trimmsignal
           *************************************/
    
                
           getPWM_Input = pulseIn(PinIn13, HIGH);     //Einlesen PWM Signal
    
             #ifdef DEBUG_PWM        // Serial output on 
             Serial.print("PWM Minimal: ");
             Serial.print(setPWM_Minimal);  
             Serial.print(" => ");         
             Serial.print("PWM Input: ");
             Serial.print(getPWM_Input);
             #endif
    
                    
           if(getPWM_Input > 1500 && getPWM_Input >= setPWM_Minimal && getPWM_Input > 10)
           {
                  myservo.writeMicroseconds(getPWM_Input);
                  #ifdef DEBUG_PWM        // Serial output on 
                  getPWM_Output = getPWM_Input;
                  Serial.print(" => PWM Output: ");
                  Serial.println( getPWM_Output );
                  #endif
           }
           else if(getPWM_Input > 1500 && getPWM_Input <= setPWM_Minimal && getPWM_Input > 10)
               {
                  myservo.writeMicroseconds(setPWM_Minimal);
                  #ifdef DEBUG_PWM        // Serial output on 
                  getPWM_Output = setPWM_Minimal;
                  Serial.print(" => PWM Output: ");
                  Serial.println( getPWM_Output );
                  #endif
           }       
           else if(getPWM_Input < 1500 && getPWM_Input > 10)
           {
            
                if((3000 - getPWM_Input) > setPWM_Minimal && getPWM_Input > 10)
                {
                  myservo.writeMicroseconds((3000 - getPWM_Input));   //Berechnen des PWM Signals in Wert grösser als 1500 damit es auch bei Rückwärtsfahrt raucht nach Motordrehzahl
                  #ifdef DEBUG_PWM        // Serial output on 
                  getPWM_Output = 3000 - getPWM_Input;
                  Serial.print(" => PWM Output: ");
                  Serial.println( getPWM_Output );
                  #endif        
                }
                else if(getPWM_Input > 10)
                {
                  myservo.writeMicroseconds(setPWM_Minimal);
                  #ifdef DEBUG_PWM        // Serial output on 
                  getPWM_Output = setPWM_Minimal;
                  Serial.print(" => PWM Output: ");
                  Serial.println( getPWM_Output );
                  #endif
                }
              
           }
           else if(getPWM_Input > 10)
           {
              myservo.writeMicroseconds(setPWM_Minimal);
              #ifdef DEBUG_PWM        // Serial output on 
              getPWM_Output = setPWM_Minimal;
              Serial.print(" => PWM Output: ");
              Serial.println( getPWM_Output );
              #endif
           }
           else
           {

              myservo.writeMicroseconds(1500);
              #ifdef DEBUG_PWM        // Serial output on 
              Serial.print(" => PWM Output: ");
              Serial.println( 1500);
              #endif
            
           }
    
    
    
    

      } 

      /********************************************
       * Ende Run Mode
       ********************************************/        


}//Ende Loop

/******************************************************************************
 * Funktion zum Schalten der Leistungsfreigabe an PinOut3
 ****************************************************************************/
void setLeistungFreigabe(bool setFreigabe)
{

          #ifdef DEBUG_FRG        // Serial output on 
         Serial.println("setLeistungFreigabe Funktion: "); 
         Serial.print("Ausgang alt: ");
         Serial.print(digitalRead(PinOut3));
         #endif  
         
  if(setFreigabe == true)
  {
    digitalWrite(PinOut3, HIGH);
    setFreigabeStatus = true;
    startZeit = millis();      

          #ifdef DEBUG_FRG        // Serial output on 
          Serial.print(" Freigabe Leistung: ");
          Serial.print(digitalRead(PinOut3));
          #endif
          
  }
  else
  {
    digitalWrite(PinOut3, LOW);
    setFreigabeStatus = false;
    aktLaufzeit = aktLaufzeit + millis() - startZeit;   


          #ifdef DEBUG_FRG        // Serial output on 
          Serial.print("Sperrung Leistung: ");
          Serial.print(digitalRead(PinOut3));
          Serial.print(" | Akt. Laufzeit: ");
          Serial.println(aktLaufzeit/1000);
          #endif
          
  }

  #ifdef DEBUG_FRG        // Serial output on 
  Serial.println("");
  #endif
  

}

/******************************************************************************
 * Deaktivierung des Leistungsausgangs wenn die max. Laufzeit überschritten ist.
 ****************************************************************************/
void setDeaktLeAusgang()
{
  
  digitalWrite(PinOut3, LOW);
  setFreigabeLeistungTime = false;

}

/******************************************************************************
 * Kontrolle der max. Laufzeit
 ****************************************************************************/
void LaufzeitKontrolle()
{

  //Laufzeit Aktualisierung
  if(setFreigabeStatus == true)
  {
    aktLaufzeit = aktLaufzeit + millis() - startZeit;  
    startZeit = millis();
  }
  
  
  if(aktLaufzeit >= maxTime)
    setDeaktLeAusgang();
    
}



   
